import os.path
import unittest

import numpy as np

import kastore as kas

"""
Tests reading in the standard test files.
"""


class StandardFilesMixin:
    """
    Read in the standard files.
    """

    @classmethod
    def setUpClass(cls):
        # Figure out where this is being run from and set the test data
        # path accordingly.
        cwd = os.getcwd()
        cls.test_data_path = "test-data"
        if cwd.endswith("python"):
            cls.test_data_path = "../test-data"

    def read_file(self, filename):
        full_path = os.path.join(self.test_data_path, filename)
        return kas.load(full_path, engine=self.engine, read_all=False)

    def test_empty_file(self):
        self.assertRaises(EOFError, self.read_file, "malformed/empty_file.kas")

    def test_bad_type(self):
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_type_11.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_type_20.kas"
        )

    def test_bad_filesizes(self):
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_filesize_0_-1.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_filesize_0_1.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_filesize_0_1024.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_filesize_10_-1.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_filesize_10_1.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_filesize_10_1024.kas"
        )

    def test_bad_magic_number(self):
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_magic_number.kas"
        )

    def test_version_0(self):
        self.assertRaises(
            kas.VersionTooOldError, self.read_file, "malformed/version_0.kas"
        )

    def test_version_100(self):
        self.assertRaises(
            kas.VersionTooNewError, self.read_file, "malformed/version_100.kas"
        )

    def test_truncated_file(self):
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/truncated_file.kas"
        )

    def test_key_offset_outside_file(self):
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/key_offset_outside_file.kas"
        )

    def test_array_offset_outside_file(self):
        self.assertRaises(
            kas.FileFormatError,
            self.read_file,
            "malformed/array_offset_outside_file.kas",
        )

    def test_key_len_outside_file(self):
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/key_len_outside_file.kas"
        )

    def test_array_len_outside_file(self):
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/array_len_outside_file.kas"
        )

    def test_bad_array_start(self):
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_array_start_-8.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_array_start_-1.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_array_start_1.kas"
        )
        self.assertRaises(
            kas.FileFormatError, self.read_file, "malformed/bad_array_start_8.kas"
        )

    def test_truncated_file_correct_size(self):
        self.assertRaises(
            kas.FileFormatError,
            self.read_file,
            "malformed/truncated_file_correct_size_100.kas",
        )
        self.assertRaises(
            kas.FileFormatError,
            self.read_file,
            "malformed/truncated_file_correct_size_128.kas",
        )
        self.assertRaises(
            kas.FileFormatError,
            self.read_file,
            "malformed/truncated_file_correct_size_129.kas",
        )
        self.assertRaises(
            kas.FileFormatError,
            self.read_file,
            "malformed/truncated_file_correct_size_200.kas",
        )

    def test_all_types(self):
        dtypes = [
            "int8",
            "uint8",
            "uint32",
            "int32",
            "uint64",
            "int64",
            "float32",
            "float64",
        ]
        for n in range(10):
            filename = f"v1/all_types_{n}_elements.kas"
            data = self.read_file(filename)
            for dtype in dtypes:
                self.assertTrue(np.array_equal(data[dtype], np.arange(n, dtype=dtype)))


class TestStandardFilesPyEngine(StandardFilesMixin, unittest.TestCase):
    engine = kas.PY_ENGINE
    read_all = False


class TestStandardFilesCEngine(StandardFilesMixin, unittest.TestCase):
    engine = kas.C_ENGINE
    read_all = False


class TestStandardFilesPyEngineReadAll(StandardFilesMixin, unittest.TestCase):
    engine = kas.PY_ENGINE
    read_all = True


class TestStandardFilesCEngineReadAll(StandardFilesMixin, unittest.TestCase):
    engine = kas.C_ENGINE
    read_all = True


class TestStandardFilesLoads(StandardFilesMixin, unittest.TestCase):
    def read_file(self, filename):
        full_path = os.path.join(self.test_data_path, filename)
        with open(full_path, "rb") as f:
            encoded = f.read()
        return kas.loads(encoded)



import os
import tempfile
import unittest

import hypothesis
import hypothesis.extra.numpy as hnp
import hypothesis.strategies as hst
import numpy as np

import kastore as kas


"""
Basic tests for the storage integrity of the data.
"""


# Set the deadline to None to avoid weird behaviour on CI.
hypothesis.settings.register_profile(
    "kastore_defaults",
    deadline=None,
    # Supress warnings resultsing from inheritance
    suppress_health_check=(hypothesis.HealthCheck.differing_executors,),
)
hypothesis.settings.load_profile("kastore_defaults")

# Exclude any 'other' unicode categories:
# http://www.unicode.org/reports/tr44/#General_Category_Values
key_alphabet = hst.characters(blacklist_categories=("C",))


class TestInputErrors(unittest.TestCase):
    """
    Simple input errors.
    """

    def setUp(self):
        fd, path = tempfile.mkstemp(prefix="kas_test_input")
        os.close(fd)
        self.temp_file = path

    def tearDown(self):
        os.unlink(self.temp_file)

    def test_empty_key(self):
        self.assertRaises(ValueError, kas.dump, {"": np.zeros(1)}, self.temp_file)

    def test_2d_array(self):
        self.assertRaises(ValueError, kas.dump, {"a": np.zeros((1, 1))}, self.temp_file)

    def test_3d_array(self):
        self.assertRaises(
            ValueError, kas.dump, {"a": np.zeros((2, 2, 1))}, self.temp_file
        )


class TestRoundTrip(unittest.TestCase):
    """
    Tests that we can round trip data through a temporary file.
    """

    def setUp(self):
        fd, path = tempfile.mkstemp(prefix="kas_test_rt")
        os.close(fd)
        self.temp_file = path

    def tearDown(self):
        os.unlink(self.temp_file)

    def verify(self, data):
        for engine in [kas.C_ENGINE, kas.PY_ENGINE]:
            for read_all in [True, False]:
                kas.dump(data, self.temp_file, engine=engine)
                new_data = kas.load(self.temp_file, read_all=read_all, engine=engine)
                self.assertEqual(sorted(new_data.keys()), sorted(data.keys()))
                for key, source_array in data.items():
                    dest_array = new_data[key]
                    # Numpy's testing assert_equal will deal correctly with NaNs.
                    np.testing.assert_equal(source_array, dest_array)
                # Make sure the file is closed before opening it again.
                del new_data


class BytesRoundTripMixin:
    """
    A mixin to which can be used to inherit the same tests defined for subclasses
    of TestRoundTrip. Note this *must* be the first class inherited from.
    """

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def verify(self, data):
        encoded = kas.dumps(data)
        new_data = kas.loads(encoded)
        self.assertEqual(sorted(new_data.keys()), sorted(data.keys()))
        for key, source_array in data.items():
            dest_array = new_data[key]
            # Numpy's testing assert_equal will deal correctly with NaNs.
            np.testing.assert_equal(source_array, dest_array)


class TestRoundTripSimple(TestRoundTrip):
    """
    Simple round-trip tests for some hand crafted cases.
    """

    def test_zero_keys(self):
        self.verify({})

    def test_single_key(self):
        self.verify({"a": np.zeros(1)})

    def test_many_keys(self):
        data = {}
        for j in range(1):
            data[str(j)] = j + np.zeros(j, dtype=np.uint32)
        self.verify(data)

    def test_all_dtypes(self):
        dtypes = [
            "int8",
            "uint8",
            "uint16",
            "int16",
            "uint32",
            "int32",
            "uint64",
            "int64",
            "float32",
            "float64",
        ]
        for n in range(10):
            data = {dtype: np.arange(n, dtype=dtype) for dtype in dtypes}
            self.verify(data)


class TestRoundTripSimpleBytes(BytesRoundTripMixin, TestRoundTripSimple):
    pass


class TestRoundTripKeys(TestRoundTrip):
    """
    Test round tripping with keys generated by hypothesis.
    """

    @hypothesis.given(key=hst.text(alphabet=key_alphabet, min_size=1))
    def test_single_key(self, key):
        self.verify({key: np.zeros(1)})

    @hypothesis.given(
        keys=hst.sets(hst.text(alphabet=key_alphabet, min_size=1), min_size=1)
    )
    def test_many_keys(self, keys):
        data = {key: np.ones(j) * j for j, key in enumerate(keys)}
        self.verify(data)


class TestRoundTripKeysBytes(BytesRoundTripMixin, TestRoundTripKeys):
    pass


shape_strategy = hnp.array_shapes(max_dims=1)


class TestRoundTripDataTypes(TestRoundTrip):
    """
    Test round tripping of the various types using Hypothesis.
    """

    @hypothesis.given(value=hnp.arrays(dtype=np.uint8, shape=shape_strategy))
    def test_single_uint8(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.int8, shape=shape_strategy))
    def test_single_int8(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.int16, shape=shape_strategy))
    def test_single_int16(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.uint16, shape=shape_strategy))
    def test_single_uint16(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.int32, shape=shape_strategy))
    def test_single_int32(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.uint32, shape=shape_strategy))
    def test_single_uint32(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.int64, shape=shape_strategy))
    def test_single_int64(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.uint64, shape=shape_strategy))
    def test_single_uint64(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.float32, shape=shape_strategy))
    def test_single_float32(self, value):
        self.verify({"a": value})

    @hypothesis.given(value=hnp.arrays(dtype=np.float64, shape=shape_strategy))
    def test_single_float64(self, value):
        self.verify({"a": value})


class TestRoundTripDataTypesBytes(BytesRoundTripMixin, TestRoundTripDataTypes):
    pass



import os
import pathlib
import struct
import tempfile
import unittest

import hypothesis
import hypothesis.strategies as hst
import numpy as np

import kastore as kas
import kastore.store as store

"""
Tests checking that the file format is as it should be.
"""

# Set the deadline to None to avoid weird behaviour on CI.
hypothesis.settings.register_profile(
    "kastore_defaults",
    deadline=None,
    # Supress warnings resultsing from inheritance
    suppress_health_check=(hypothesis.HealthCheck.differing_executors,),
)
hypothesis.settings.load_profile("kastore_defaults")


# Exclude any 'other' unicode categories:
# http://www.unicode.org/reports/tr44/#General_Category_Values
key_alphabet = hst.characters(blacklist_categories=("C",))


class TestFileSignature(unittest.TestCase):
    """
    Checks the file signature is what we think it should be.
    """

    def test_form(self):
        self.assertEqual(len(store.MAGIC), 8)
        self.assertEqual(b"\211KAS\r\n\032\n", store.MAGIC)


class FormatMixin:
    """
    Tests for the file format.
    """

    def setUp(self):
        fd, self.temp_file = tempfile.mkstemp(suffix=".kas", prefix="kas_rt_test")
        os.close(fd)

    def tearDown(self):
        os.unlink(self.temp_file)

    def test_header_format(self):
        for n in range(10):
            kas.dump(
                {str(j): np.zeros(1) for j in range(n)},
                self.temp_file,
                engine=self.engine,
            )
            with open(self.temp_file, "rb") as f:
                contents = f.read()
            self.assertEqual(contents[0:8], store.MAGIC)
            major, minor, num_items, size = struct.unpack("<HHIQ", contents[8:24])
            self.assertEqual(major, store.VERSION_MAJOR)
            self.assertEqual(minor, store.VERSION_MINOR)
            self.assertEqual(num_items, n)
            self.assertEqual(size, len(contents))
            trailer = contents[24 : store.HEADER_SIZE]
            # The remainder should be zeros.
            self.assertEqual(
                trailer, bytearray(0 for _ in range(store.HEADER_SIZE - 24))
            )

    def test_zero_items(self):
        kas.dump({}, self.temp_file, engine=self.engine)
        with open(self.temp_file, "rb") as f:
            contents = f.read()
        self.assertEqual(len(contents), 64)

    def test_item_descriptor_format(self):
        for n in range(10):
            kas.dump(
                {str(j): j * np.ones(j) for j in range(n)},
                self.temp_file,
                engine=self.engine,
            )
            with open(self.temp_file, "rb") as f:
                contents = f.read()
            self.assertEqual(struct.unpack("<I", contents[12:16])[0], n)
            offset = store.HEADER_SIZE
            for _ in range(n):
                descriptor = contents[offset : offset + store.ITEM_DESCRIPTOR_SIZE]
                offset += store.ITEM_DESCRIPTOR_SIZE
                type_ = struct.unpack("<B", descriptor[0:1])[0]
                key_start, key_len, array_start, array_len = struct.unpack(
                    "<QQQQ", descriptor[8:40]
                )
                trailer = descriptor[40 : store.ITEM_DESCRIPTOR_SIZE]
                # The remainder should be zeros.
                self.assertEqual(
                    trailer,
                    bytearray(0 for _ in range(store.ITEM_DESCRIPTOR_SIZE - 40)),
                )
                self.assertEqual(descriptor[1:4], bytearray([0, 0, 0]))
                self.assertEqual(type_, store.FLOAT64)
                self.assertGreater(key_start, 0)
                self.assertGreater(key_len, 0)
                self.assertGreater(array_start, 0)
                self.assertGreaterEqual(array_len, 0)

    def validate_storage(self, data):
        kas.dump(data, self.temp_file, engine=self.engine)
        with open(self.temp_file, "rb") as f:
            contents = f.read()
        offset = store.HEADER_SIZE
        descriptors = []
        for _ in range(len(data)):
            descriptor = store.ItemDescriptor.unpack(
                contents[offset : offset + store.ItemDescriptor.size]
            )
            descriptors.append(descriptor)
            offset += store.ItemDescriptor.size
        # Keys must be sorted lexicographically.
        sorted_keys = sorted(data.keys())
        # Keys should be packed sequentially immediately after the descriptors.
        offset = store.HEADER_SIZE + len(data) * store.ITEM_DESCRIPTOR_SIZE
        for d, key in zip(descriptors, sorted_keys):
            self.assertEqual(d.key_start, offset)
            unpacked_key = contents[d.key_start : d.key_start + d.key_len]
            self.assertEqual(key.encode("utf8"), unpacked_key)
            offset += d.key_len
        # Arrays should be packed sequentially immediately after the keys on
        # 8 byte boundaries
        for d, key in zip(descriptors, sorted_keys):
            remainder = offset % 8
            if remainder != 0:
                offset += 8 - remainder
            self.assertEqual(d.array_start, offset)
            nbytes = d.array_len * store.type_size(d.type)
            array = np.frombuffer(
                contents[d.array_start : d.array_start + nbytes],
                dtype=store.type_to_np_dtype_map[d.type],
            )
            np.testing.assert_equal(data[key], array)
            offset += nbytes

    def test_simple_key_storage(self):
        for n in range(10):
            self.validate_storage({"a" * (j + 1): np.ones(1) for j in range(n)})

    def test_simple_array_storage(self):
        for n in range(10):
            self.validate_storage({str(j): j * np.ones(j) for j in range(n)})

    # Note that hypothesis seems to be leaking memory, so when we're running tests
    # against the C API for memory leaks this must be commented out.
    @hypothesis.given(
        keys=hst.sets(hst.text(alphabet=key_alphabet, min_size=1), min_size=1)
    )
    def test_many_keys(self, keys):
        data = {key: np.ones(j, dtype=np.int32) * j for j, key in enumerate(keys)}
        self.validate_storage(data)


class TestFormatPyEngine(FormatMixin, unittest.TestCase):
    engine = kas.PY_ENGINE


class TestFormatCEngine(FormatMixin, unittest.TestCase):
    engine = kas.C_ENGINE


class TestAlignedPacking(unittest.TestCase):
    """
    Tests that we are correctly computing alignments for the arrays.
    """

    def test_descriptor_str(self):
        items = {"a": np.array([1], dtype=np.int8)}
        descriptors, file_size = store.pack_items(items)
        self.assertGreater(len(str(descriptors[0])), 0)

    def test_single_key_1(self):
        items = {"a": np.array([1], dtype=np.int8)}
        descriptors, file_size = store.pack_items(items)
        d = descriptors[0]
        self.assertEqual(d.key_start, 128)
        self.assertEqual(d.key_len, 1)
        self.assertEqual(d.array_start, 136)
        self.assertEqual(d.array_len, 1)

    def test_single_key_7(self):
        items = {"aaaaaaa": np.array([1], dtype=np.int8)}
        descriptors, file_size = store.pack_items(items)
        d = descriptors[0]
        self.assertEqual(d.key_start, 128)
        self.assertEqual(d.key_len, 7)
        self.assertEqual(d.array_start, 136)
        self.assertEqual(d.array_len, 1)

    def test_single_key_8(self):
        items = {"aaaaaaaa": np.array([1], dtype=np.int8)}
        descriptors, file_size = store.pack_items(items)
        d = descriptors[0]
        self.assertEqual(d.key_start, 128)
        self.assertEqual(d.key_len, 8)
        self.assertEqual(d.array_start, 136)
        self.assertEqual(d.array_len, 1)

    def test_two_keys_array_len1(self):
        a = np.array([1], dtype=np.int8)
        items = {"a": a, "b": a}
        descriptors, file_size = store.pack_items(items)
        d = descriptors[0]
        self.assertEqual(d.key_start, 192)
        self.assertEqual(d.key_len, 1)
        self.assertEqual(d.array_start, 200)
        self.assertEqual(d.array_len, 1)
        d = descriptors[1]
        self.assertEqual(d.key_start, 193)
        self.assertEqual(d.key_len, 1)
        self.assertEqual(d.array_start, 208)
        self.assertEqual(d.array_len, 1)

    def test_two_keys_array_len8(self):
        a = np.array([1], dtype=np.int64)
        items = {"a": a, "b": a}
        descriptors, file_size = store.pack_items(items)
        d = descriptors[0]
        self.assertEqual(d.key_start, 192)
        self.assertEqual(d.key_len, 1)
        self.assertEqual(d.array_start, 200)
        self.assertEqual(d.array_len, 1)
        d = descriptors[1]
        self.assertEqual(d.key_start, 193)
        self.assertEqual(d.key_len, 1)
        self.assertEqual(d.array_start, 208)
        self.assertEqual(d.array_len, 1)

    def test_two_keys_array_len4(self):
        a = np.array([1], dtype=np.int32)
        items = {"a": a, "b": a}
        descriptors, file_size = store.pack_items(items)
        d = descriptors[0]
        self.assertEqual(d.key_start, 192)
        self.assertEqual(d.key_len, 1)
        self.assertEqual(d.array_start, 200)
        self.assertEqual(d.array_len, 1)
        d = descriptors[1]
        self.assertEqual(d.key_start, 193)
        self.assertEqual(d.key_len, 1)
        self.assertEqual(d.array_start, 208)
        self.assertEqual(d.array_len, 1)


class TestEnginesProduceIdenticalFiles(unittest.TestCase):
    """
    Ensure that the two engines produces identical files.
    """

    def setUp(self):
        fd, self.temp_file = tempfile.mkstemp(suffix=".kas", prefix="kas_identity_test")
        os.close(fd)

    def tearDown(self):
        os.unlink(self.temp_file)

    def verify(self, data):
        kas.dump(data, self.temp_file, engine=kas.C_ENGINE)
        with open(self.temp_file, "rb") as f:
            c_file = f.read()
        kas.dump(data, self.temp_file, engine=kas.PY_ENGINE)
        with open(self.temp_file, "rb") as f:
            py_file = f.read()
        self.assertEqual(c_file, py_file)

    def test_empty(self):
        self.verify({})

    def test_one_key_empty_array(self):
        self.verify({"a": []})

    def test_many_keys_empty_arrays(self):
        self.verify({"a" * (j + 1): [] for j in range(10)})

    def test_many_keys_nonempty_arrays(self):
        for dtype in [np.int8, np.uint8, np.uint32, np.float64]:
            self.verify({"a" * (j + 1): np.arange(j, dtype=dtype) for j in range(10)})

    def verify_all_dtypes(self, n):
        dtypes = [
            "int8",
            "uint8",
            "int16",
            "uint16",
            "uint32",
            "int32",
            "uint64",
            "int64",
            "float32",
            "float64",
        ]
        data = {dtype: np.arange(n, dtype=dtype) for dtype in dtypes}
        self.verify(data)

    def test_all_dtypes_0_elements(self):
        self.verify_all_dtypes(0)

    def test_all_dtypes_1_elements(self):
        self.verify_all_dtypes(1)

    def test_all_dtypes_2_elements(self):
        self.verify_all_dtypes(2)

    def test_all_dtypes_3_elements(self):
        self.verify_all_dtypes(3)

    def test_all_dtypes_4_elements(self):
        self.verify_all_dtypes(4)


class TruncatedFilesMixin:
    """
    Tests that we return the correct errors when we have truncated files.
    """

    def test_zero_bytes(self):
        with tempfile.TemporaryDirectory() as tempdir:
            path = pathlib.Path(tempdir) / "testfile"
            with open(path, "wb") as f:
                f.write(b"")
            self.assertTrue(path.exists())
            with self.assertRaises(EOFError):
                kas.load(path, engine=self.engine)

    def test_short_file(self):
        example = {"a": np.arange(2)}
        encoded = kas.dumps(example)
        with tempfile.TemporaryDirectory() as tempdir:
            path = pathlib.Path(tempdir) / "testfile"
            for j in range(1, 64):  # len(encoded) - 2):
                with open(path, "wb") as f:
                    f.write(encoded[:j])
                self.assertTrue(path.exists())
                with self.assertRaises(kas.FileFormatError):
                    kas.load(path, engine=self.engine)


class TestTruncatedFilesPyEngine(unittest.TestCase, TruncatedFilesMixin):
    engine = kas.PY_ENGINE


class TestTruncatedFilesCEngine(unittest.TestCase, TruncatedFilesMixin):
    engine = kas.C_ENGINE



import os
import platform
import struct
import tempfile
import unittest

import numpy as np

import kastore as kas
import kastore.store as store

"""
Tests for error conditions.
"""

IS_WINDOWS = platform.system() == "Windows"


class InterfaceMixin:
    """
    Exercise the low-level interface.
    """

    def setUp(self):
        fd, path = tempfile.mkstemp(prefix="kas_test_errors")
        os.close(fd)
        self.temp_file = path

    def tearDown(self):
        os.unlink(self.temp_file)

    def test_bad_dicts(self):
        for bad_dict in [[], "w34", None, 1]:
            self.assertRaises(
                TypeError, kas.dump, bad_dict, self.temp_file, engine=self.engine
            )
            self.assertRaises(
                TypeError, kas.dump, bad_dict, self.temp_file, engine=self.engine
            )

    def test_bad_filename_type(self):
        for bad_filename in [[], None, {}]:
            self.assertRaises(TypeError, kas.dump, {}, bad_filename, engine=self.engine)
            self.assertRaises(TypeError, kas.dump, {}, bad_filename, engine=self.engine)
            self.assertRaises(TypeError, kas.load, bad_filename, engine=self.engine)
            self.assertRaises(TypeError, kas.load, bad_filename, engine=self.engine)

    def test_bad_keys(self):
        a = np.zeros(1)
        for bad_key in [(1234,), b"1234", None, 1234]:
            with self.assertRaises(TypeError):
                kas.dump({bad_key: a}, self.temp_file, engine=self.engine)

    def test_bad_arrays(self):
        kas.dump({"a": []}, self.temp_file, engine=self.engine)
        for bad_array in [kas, lambda x: x, "1234", None, [[0, 1], [0, 2]]]:
            self.assertRaises(
                ValueError,
                kas.dump,
                {"a": bad_array},
                self.temp_file,
                engine=self.engine,
            )
        # TODO add tests for arrays in fortran order and so on.

    def test_file_not_found(self):
        a = np.zeros(1)
        for bad_file in ["no_such_file", "/no/such/file"]:
            self.assertRaises(FileNotFoundError, kas.load, bad_file, engine=self.engine)
        self.assertRaises(
            FileNotFoundError, kas.dump, {"a": a}, "/no/such/file", engine=self.engine
        )

    def test_file_is_a_directory(self):
        tmp_dir = tempfile.mkdtemp()
        try:
            exception = IsADirectoryError
            if IS_WINDOWS:
                exception = PermissionError
            self.assertRaises(
                exception, kas.dump, {"a": []}, tmp_dir, engine=self.engine
            )
            self.assertRaises(exception, kas.load, tmp_dir, engine=self.engine)
        finally:
            os.rmdir(tmp_dir)


class TestInterfacePyEngine(InterfaceMixin, unittest.TestCase):
    engine = kas.PY_ENGINE


class TestInterfaceCEngine(InterfaceMixin, unittest.TestCase):
    engine = kas.C_ENGINE


class TestEngines(unittest.TestCase):
    """
    Check that we correctly identify bad engines
    """

    bad_engines = [None, {}, "no such engine", b"not an engine"]

    def test_bad_engine_dump(self):
        for bad_engine in self.bad_engines:
            self.assertRaises(ValueError, kas.dump, {}, "", engine=bad_engine)

    def test_bad_engine_load(self):
        for bad_engine in self.bad_engines:
            self.assertRaises(ValueError, kas.load, "", engine=bad_engine)


class FileFormatsMixin:
    """
    Common utilities for tests on the file format.
    """

    def setUp(self):
        fd, path = tempfile.mkstemp(prefix="kas_malformed_files")
        os.close(fd)
        self.temp_file = path

    def tearDown(self):
        os.unlink(self.temp_file)

    def write_file(self, num_items=0):
        data = {}
        for j in range(num_items):
            data["a" * (j + 1)] = np.arange(j)
        kas.dump(data, self.temp_file)


class MalformedFilesMixin(FileFormatsMixin):
    """
    Tests for various types of malformed intput files.
    """

    def test_empty_file(self):
        with open(self.temp_file, "w"):
            pass
        self.assertEqual(os.path.getsize(self.temp_file), 0)
        self.assertRaises(
            EOFError,
            kas.load,
            self.temp_file,
            engine=self.engine,
            read_all=self.read_all,
        )

    def test_bad_magic(self):
        self.write_file()
        with open(self.temp_file, "rb") as f:
            buff = bytearray(f.read())
        before_len = len(buff)
        buff[0:8] = b"12345678"
        self.assertEqual(len(buff), before_len)
        with open(self.temp_file, "wb") as f:
            f.write(buff)
        self.assertRaises(
            kas.FileFormatError,
            kas.load,
            self.temp_file,
            engine=self.engine,
            read_all=self.read_all,
        )

    def test_bad_file_size(self):
        for num_items in range(10):
            for offset in [-2, -1, 1, 2**10]:
                self.write_file(num_items)
                file_size = os.path.getsize(self.temp_file)
                with open(self.temp_file, "rb") as f:
                    buff = bytearray(f.read())
                before_len = len(buff)
                buff[16:24] = struct.pack("<Q", file_size + offset)
                self.assertEqual(len(buff), before_len)
                with open(self.temp_file, "wb") as f:
                    f.write(buff)
                with self.assertRaises(kas.FileFormatError):
                    kas.load(self.temp_file, engine=self.engine, read_all=self.read_all)

    def test_truncated_file_descriptors(self):
        for num_items in range(2, 5):
            self.write_file(num_items)
            with open(self.temp_file, "rb") as f:
                buff = bytearray(f.read())
            with open(self.temp_file, "wb") as f:
                f.write(buff[: num_items * store.ITEM_DESCRIPTOR_SIZE - 1])
            with self.assertRaises(kas.FileFormatError):
                kas.load(self.temp_file, engine=self.engine, read_all=self.read_all)

    def test_truncated_file_data(self):
        for num_items in range(2, 5):
            self.write_file(num_items)
            with open(self.temp_file, "rb") as f:
                buff = bytearray(f.read())
            with open(self.temp_file, "wb") as f:
                f.write(buff[:-1])
            with self.assertRaises(kas.FileFormatError):
                # Must call dict to ensure all the keys are loaded.
                dict(
                    kas.load(self.temp_file, engine=self.engine, read_all=self.read_all)
                )

    def test_bad_item_types(self):
        items = {"a": []}
        descriptors, file_size = store.pack_items(items)
        num_types = len(store.np_dtype_to_type_map)
        for bad_type in [num_types + 1, 2 * num_types]:
            with open(self.temp_file, "wb") as f:
                descriptors[0].type = bad_type
                store.write_file(f, descriptors, file_size)
            with self.assertRaises(kas.FileFormatError):
                kas.load(self.temp_file, engine=self.engine, read_all=self.read_all)

    def test_bad_key_initial_offsets(self):
        items = {"a": np.arange(100)}
        # First key offset must be at header_size + n * (descriptor_size)
        for offset in [-1, +1, 2, 100]:
            # First key offset must be at header_size + n * (descriptor_size)
            descriptors, file_size = store.pack_items(items)
            descriptors[0].key_start += offset
            with open(self.temp_file, "wb") as f:
                store.write_file(f, descriptors, file_size)
            self.assertRaises(
                kas.FileFormatError,
                kas.load,
                self.temp_file,
                engine=self.engine,
                read_all=self.read_all,
            )

    def test_bad_key_non_sequential(self):
        items = {"a": np.arange(100), "b": []}
        # Keys must be packed sequentially.
        for offset in [-1, +1, 2, 100]:
            descriptors, file_size = store.pack_items(items)
            descriptors[1].key_start += offset
            with open(self.temp_file, "wb") as f:
                store.write_file(f, descriptors, file_size)
            self.assertRaises(
                kas.FileFormatError,
                kas.load,
                self.temp_file,
                engine=self.engine,
                read_all=self.read_all,
            )

    def test_bad_array_initial_offset(self):
        items = {"a": np.arange(100)}
        for offset in [-100, -1, +1, 2, 8, 16, 100]:
            # First key offset must be at header_size + n * (descriptor_size)
            descriptors, file_size = store.pack_items(items)
            descriptors[0].array_start += offset
            with open(self.temp_file, "wb") as f:
                store.write_file(f, descriptors, file_size)
            self.assertRaises(
                kas.FileFormatError,
                kas.load,
                self.temp_file,
                engine=self.engine,
                read_all=self.read_all,
            )

    def test_bad_array_non_sequential(self):
        items = {"a": np.arange(100), "b": []}
        for offset in [-1, 1, 2, -8, 8, 100]:
            descriptors, file_size = store.pack_items(items)
            descriptors[1].array_start += offset
            with open(self.temp_file, "wb") as f:
                store.write_file(f, descriptors, file_size)
            self.assertRaises(
                kas.FileFormatError,
                kas.load,
                self.temp_file,
                engine=self.engine,
                read_all=self.read_all,
            )

    def test_bad_array_alignment(self):
        items = {"a": np.arange(100, dtype=np.int8), "b": []}
        descriptors, file_size = store.pack_items(items)
        descriptors[0].array_start += 1
        descriptors[0].array_len -= 1
        with open(self.temp_file, "wb") as f:
            store.write_file(f, descriptors, file_size)
        with self.assertRaises(kas.FileFormatError):
            kas.load(self.temp_file, engine=self.engine, read_all=self.read_all)

    def test_bad_array_packing(self):
        items = {"a": np.arange(100, dtype=np.int8), "b": []}
        descriptors, file_size = store.pack_items(items)
        descriptors[0].array_start += 8
        descriptors[0].array_len -= 8
        with open(self.temp_file, "wb") as f:
            store.write_file(f, descriptors, file_size)
        with self.assertRaises(kas.FileFormatError):
            kas.load(self.temp_file, engine=self.engine, read_all=self.read_all)


class TestMalformedFilesPyEngine(MalformedFilesMixin, unittest.TestCase):
    read_all = False
    engine = kas.PY_ENGINE


class TestMalformedFilesCEngine(MalformedFilesMixin, unittest.TestCase):
    read_all = False
    engine = kas.C_ENGINE


class TestMalformedFilesPyEngineReadAll(MalformedFilesMixin, unittest.TestCase):
    read_all = True
    engine = kas.PY_ENGINE


class TestMalformedFilesCEngineReadAll(MalformedFilesMixin, unittest.TestCase):
    read_all = True
    engine = kas.C_ENGINE


class FileVersionsMixin(FileFormatsMixin):
    """
    Tests for the file major version.
    """

    def verify_major_version(self, version):
        self.write_file()
        with open(self.temp_file, "rb") as f:
            buff = bytearray(f.read())
        before_len = len(buff)
        buff[8:10] = struct.pack("<H", version)
        self.assertEqual(len(buff), before_len)
        with open(self.temp_file, "wb") as f:
            f.write(buff)
        kas.load(self.temp_file, engine=self.engine, read_all=self.read_all)

    def test_major_version_too_old(self):
        self.assertRaises(
            kas.VersionTooOldError, self.verify_major_version, store.VERSION_MAJOR - 1
        )

    def test_major_version_too_new(self):
        for j in range(1, 5):
            self.assertRaises(
                kas.VersionTooNewError,
                self.verify_major_version,
                store.VERSION_MAJOR + j,
            )


class TestFileVersionsPyEngine(FileVersionsMixin, unittest.TestCase):
    engine = kas.PY_ENGINE
    read_all = False


class TestFileVersionsPyEngineReadAll(FileVersionsMixin, unittest.TestCase):
    engine = kas.PY_ENGINE
    read_all = True


class TestFileVersionsCEngine(FileVersionsMixin, unittest.TestCase):
    engine = kas.C_ENGINE
    read_all = False


class TestFileVersionsCEngineReadAll(FileVersionsMixin, unittest.TestCase):
    engine = kas.C_ENGINE
    read_all = True



import io
import os
import pathlib
import tempfile
import unittest

import numpy as np

import kastore as kas
import kastore.exceptions as exceptions

"""
Basic tests for the information API.
"""


class InterfaceTest(unittest.TestCase):
    """
    Superclass of tests that assess the kastore module interface.
    """

    def setUp(self):
        fd, path = tempfile.mkstemp(prefix="kas_test_info")
        os.close(fd)
        self.temp_file = path

    def tearDown(self):
        try:
            os.unlink(self.temp_file)
        except OSError:
            pass

    def verify_dicts_equal(self, d1, d2):
        self.assertEqual(sorted(d1.keys()), sorted(d2.keys()))
        for key in d1.keys():
            np.testing.assert_equal(d1[key], d2[key])


class TestBasicInfo(InterfaceTest):
    """
    Check that the info we return is accurate.
    """

    def verify(self, data):
        kas.dump(data, self.temp_file)
        for read_all in [True, False]:
            new_data = kas.load(self.temp_file, read_all=read_all)
            for key, array in new_data.items():
                info = new_data.info(key)
                s = str(info)
                self.assertGreater(len(s), 0)
                self.assertEqual(array.nbytes, info.size)
                self.assertEqual(array.shape, info.shape)
                self.assertEqual(array.dtype, np.dtype(info.dtype))

    def test_all_dtypes(self):
        dtypes = [
            "int8",
            "uint8",
            "uint32",
            "int32",
            "uint64",
            "int64",
            "float32",
            "float64",
        ]
        for n in range(10):
            data = {dtype: np.arange(n, dtype=dtype) for dtype in dtypes}
            self.verify(data)


class TestClosedStore(InterfaceTest):
    """
    Checks that a closed store is no longer accessible.
    """

    def verify_closed(self, store):
        self.assertRaises(exceptions.StoreClosedError, store.get, "a")
        self.assertRaises(exceptions.StoreClosedError, store.info, "a")
        self.assertRaises(exceptions.StoreClosedError, list, store.keys())
        self.assertRaises(exceptions.StoreClosedError, list, store.items())

    def test_context_manager(self):
        N = 100
        data = {"a": np.arange(N)}
        kas.dump(data, self.temp_file)
        with kas.load(self.temp_file) as store:
            self.assertIn("a", store)
            self.assertTrue(np.array_equal(store["a"], np.arange(N)))
        self.verify_closed(store)

    def test_manual_close(self):
        N = 100
        data = {"a": np.arange(N)}
        kas.dump(data, self.temp_file)
        store = kas.load(self.temp_file)
        self.assertIn("a", store)
        self.assertTrue(np.array_equal(store["a"], np.arange(N)))
        store.close()
        self.verify_closed(store)


class StoreRoundTripMixin:
    """
    We should be able to feed the output of load directly to dump.
    """

    def verify(self, data):
        kas.dump(data, self.temp_file, engine=self.engine)
        copy = kas.load(self.temp_file, engine=self.engine)
        kas.dump(copy, self.temp_file, engine=self.engine)
        copy = kas.load(self.temp_file, engine=self.engine)
        self.verify_dicts_equal(copy, data)

    def test_empty(self):
        self.verify({})

    def test_non_empty(self):
        self.verify({"a": np.arange(10), "b": np.zeros(100)})


class TestStoreRoundTripPyEngine(StoreRoundTripMixin, InterfaceTest):
    engine = kas.PY_ENGINE


class TestStoreRoundTripCEngine(StoreRoundTripMixin, InterfaceTest):
    engine = kas.C_ENGINE


class TestBytesIOInput(InterfaceTest):
    """
    Tests that we get the expected behaviour when using BytesIO.
    """

    def test_py_engine_single(self):
        data = {"a": np.arange(10), "b": np.zeros(100)}
        fileobj = io.BytesIO()
        kas.dump(data, fileobj, engine=kas.PY_ENGINE)
        fileobj.seek(0)
        data_2 = kas.load(fileobj, engine=kas.PY_ENGINE)
        self.verify_dicts_equal(data, data_2)

    def test_py_engine_multi(self):
        data = {"a": np.arange(10), "b": np.zeros(100)}
        n = 10
        fileobj = io.BytesIO()
        for _ in range(n):
            kas.dump(data, fileobj, engine=kas.PY_ENGINE)
        fileobj.seek(0)
        for _ in range(n):
            data_2 = kas.load(fileobj, read_all=True, engine=kas.PY_ENGINE)
            self.verify_dicts_equal(data, data_2)

    def test_c_engine_fails(self):
        data = {"a": np.arange(10), "b": np.zeros(100)}
        fileobj = io.BytesIO()
        with self.assertRaises(io.UnsupportedOperation):
            kas.dump(data, fileobj, engine=kas.C_ENGINE)
        with self.assertRaises(io.UnsupportedOperation):
            kas.load(fileobj, engine=kas.C_ENGINE)


class TestGetInclude(unittest.TestCase):
    """
    Check that the get_include works as expected.
    """

    def test_output(self):
        include_dir = kas.get_include()
        self.assertTrue(os.path.exists(include_dir))
        self.assertTrue(os.path.isdir(include_dir))
        path = os.path.join(kas.__path__[0], "include")
        self.assertEqual(include_dir, os.path.abspath(path))


class TestMissingCEngine(InterfaceTest):
    """
    Tests that we handle the missing low level module smoothly.
    """

    def test_dump(self):
        data = {"a": np.zeros(1)}
        try:
            kas._kastore_loaded = False
            with self.assertRaises(RuntimeError):
                kas.dump(data, self.temp_file, engine=kas.C_ENGINE)
        finally:
            kas._kastore_loaded = True

    def test_load(self):
        data = {"a": np.zeros(1)}
        kas.dump(data, self.temp_file)
        try:
            kas._kastore_loaded = False
            with self.assertRaises(RuntimeError):
                kas.load(self.temp_file, engine=kas.C_ENGINE)
        finally:
            kas._kastore_loaded = True


class TestOpenSemantics(unittest.TestCase):
    """
    Tests that we can open file-like objects with the correct semantics.
    """

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory()

    def tearDown(self):
        self.temp_dir.cleanup()

    def verify_path(self, path):
        with kas._open_file(path, "w") as f:
            f.write("xyz")
        with open(path) as f:
            self.assertEqual(f.read(), "xyz")

    def test_string(self):
        path = os.path.join(self.temp_dir.name, "testfile")
        self.verify_path(path)

    def test_bytes(self):
        path = os.path.join(self.temp_dir.name, "testfile")
        self.verify_path(path.encode())

    def test_pathlib(self):
        path = pathlib.Path(self.temp_dir.name) / "testfile"
        self.verify_path(path)

    def test_fd(self):
        path = os.path.join(self.temp_dir.name, "testfile")
        with open(path, "wb") as f:
            fd = f.fileno()
            stat_before = os.fstat(fd)
            with kas._open_file(fd, "wb") as f:
                f.write(b"xyz")
            stat_after = os.fstat(fd)
            self.assertEqual(stat_before.st_mode, stat_after.st_mode)
        with open(path, "rb") as f:
            self.assertEqual(f.read(), b"xyz")

    def test_regular_file(self):
        path = os.path.join(self.temp_dir.name, "testfile")
        with open(path, "w") as f_input:
            with kas._open_file(f_input, "w") as f:
                f.write("xyz")
        with open(path) as f:
            self.assertEqual(f.read(), "xyz")

    def test_temp_file(self):
        with tempfile.TemporaryFile() as f_input:
            with kas._open_file(f_input, "wb") as f:
                f.write(b"xyz")
            f_input.seek(0)
            self.assertEqual(f_input.read(), b"xyz")

    def test_string_io(self):
        buff = io.StringIO()
        with kas._open_file(buff, "w") as f:
            f.write("xyz")
        self.assertEqual(buff.getvalue(), "xyz")

    def test_bytes_io(self):
        buff = io.BytesIO()
        with kas._open_file(buff, "wb") as f:
            f.write(b"xyz")
        self.assertEqual(buff.getvalue(), b"xyz")

    def test_non_file(self):
        for bad_type in [None, {}, []]:
            with self.assertRaises(TypeError):
                with kas._open_file(bad_type, "wb"):
                    pass


class TestExceptions(unittest.TestCase):
    def test_inheritance(self):
        # Make sure all our exceptions are subclasses of KastoreException.
        exceptions = [
            kas.FileFormatError,
            kas.VersionTooNewError,
            kas.VersionTooOldError,
            kas.StoreClosedError,
        ]
        for exc_class in exceptions:
            self.assertTrue(issubclass(exc_class, kas.KastoreException))

    def test_format_error(self):
        with self.assertRaises(kas.FileFormatError):
            kas.loads(b"x" * 1024)
        # This is also a KastoreException
        with self.assertRaises(kas.KastoreException):
            kas.loads(b"x" * 1024)



import os
import platform
import tempfile
import unittest

import numpy as np

import _kastore

"""
Tests for the low-level C interface
"""


IS_WINDOWS = platform.system() == "Windows"


@unittest.skipIf(IS_WINDOWS, "Not worth making this work on windows")
class TestBasicOperation(unittest.TestCase):
    """
    Simple tests just to verify the basic operation.
    """

    def assertDataEqual(self, d1, d2):
        self.assertEqual(d1.keys(), d2.keys())
        for key in d1.keys():
            np.testing.assert_array_equal(d1[key], d2[key])

    def test_file_round_trip(self):
        with tempfile.TemporaryFile() as f:
            data = {"a": np.arange(10)}
            _kastore.dump(data, f)
            f.seek(0)
            x = _kastore.load(f)
            self.assertDataEqual(x, data)

    def test_fileno_round_trip(self):
        with tempfile.TemporaryFile() as f:
            data = {"a": np.arange(10)}
            _kastore.dump(data, f.fileno())
            f.seek(0)
            x = _kastore.load(f.fileno())
            self.assertDataEqual(x, data)

    def test_same_file_round_trip(self):
        with tempfile.TemporaryFile() as f:
            for j in range(10):
                start = f.tell()
                data = {"a": np.arange(j * 100)}
                _kastore.dump(data, f)
                f.seek(start)
                x = _kastore.load(f)
                self.assertDataEqual(x, data)


@unittest.skipIf(IS_WINDOWS, "Not worth making this work on windows")
class TestInputs(unittest.TestCase):
    def test_bad_numeric_fd(self):
        for fd in ["1", 1.0, 2.0]:
            with self.assertRaises(TypeError):
                _kastore.dump({}, fd)
            with self.assertRaises(TypeError):
                _kastore.load(fd)

    def test_bad_fd(self):
        bad_fd = 10000
        with self.assertRaises(OSError):
            _kastore.dump({}, bad_fd)
        with self.assertRaises(OSError):
            _kastore.load(bad_fd)
        bad_fd = -1
        with self.assertRaises(ValueError):
            _kastore.dump({}, bad_fd)
        with self.assertRaises(ValueError):
            _kastore.load(bad_fd)

    def test_bad_file_mode(self):
        with open(os.devnull) as f:
            with self.assertRaises(OSError):
                _kastore.dump({}, f)
        with open(os.devnull, "w") as f:
            with self.assertRaises(OSError):
                _kastore.load(f)

    def test_bad_load_args(self):
        with self.assertRaises(TypeError):
            _kastore.load(0, read_all="sdf")

    def test_bad_dtype(self):
        with open(os.devnull, "w") as f:
            with self.assertRaises(ValueError):
                # complex number
                array = np.array([0], dtype="c16")
                _kastore.dump({"a": array}, f)



import io
import logging
import os
import subprocess
import sys
import tempfile
import unittest
from unittest import mock

import numpy as np

import kastore as kas
import kastore.__main__ as main
import kastore.cli as cli

"""
Test cases for the kastore CLI.
"""


def capture_output(func, *args, **kwargs):
    """
    Runs the specified function and arguments, and returns the
    tuple (stdout, stderr) as strings.
    """
    stdout = sys.stdout
    sys.stdout = io.StringIO()
    stderr = sys.stderr
    sys.stderr = io.StringIO()

    try:
        func(*args, **kwargs)
        stdout_output = sys.stdout.getvalue()
        stderr_output = sys.stderr.getvalue()
    finally:
        sys.stdout.close()
        sys.stdout = stdout
        sys.stderr.close()
        sys.stderr = stderr
    return stdout_output, stderr_output


class TestMain(unittest.TestCase):
    """
    Simple tests for the main function.
    """

    def test_cli_main(self):
        with mock.patch("argparse.ArgumentParser.parse_args") as mocked_parse:
            cli.kastore_main()
            mocked_parse.assert_called_once()

    def test_main(self):
        with mock.patch("argparse.ArgumentParser.parse_args") as mocked_parse:
            main.main()
            mocked_parse.assert_called_once()


class TestListArgumentParser(unittest.TestCase):
    """
    Tests the parser to ensure it parses input values correctly.
    """

    def parse_args(self, args):
        parser = cli.get_kastore_parser()
        return parser.parse_args(args)

    def test_defaults(self):
        args = self.parse_args(["ls", "filename"])
        self.assertEqual(args.store, "filename")
        self.assertEqual(args.long, False)
        self.assertEqual(args.human_readable, False)

    def test_long(self):
        args = self.parse_args(["ls", "-l", "filename"])
        self.assertEqual(args.store, "filename")
        self.assertEqual(args.long, True)
        args = self.parse_args(["ls", "--long", "filename"])
        self.assertEqual(args.store, "filename")
        self.assertEqual(args.long, True)

    def test_human_readable(self):
        args = self.parse_args(["ls", "-H", "filename"])
        self.assertEqual(args.store, "filename")
        self.assertEqual(args.human_readable, True)
        args = self.parse_args(["ls", "--human-readable", "filename"])
        self.assertEqual(args.store, "filename")
        self.assertEqual(args.human_readable, True)


class TestDumpArgumentParser(unittest.TestCase):
    """
    Tests the parser to ensure it parses input values correctly.
    """

    def parse_args(self, args):
        parser = cli.get_kastore_parser()
        return parser.parse_args(args)

    def test_defaults(self):
        args = self.parse_args(["dump", "filename", "array"])
        self.assertEqual(args.store, "filename")
        self.assertEqual(args.array, "array")


class TestDirectOutput(unittest.TestCase):
    """
    Tests for some of the argparse inherited functionality
    """

    def run_command(self, cmd):
        stdout = subprocess.check_output(
            [sys.executable, "-m", "kastore"] + cmd, stderr=subprocess.STDOUT
        )
        return stdout.decode()

    def test_help(self):
        stdout = self.run_command(["-h"])
        self.assertGreater(len(stdout), 0)

    def test_version(self):
        stdout = self.run_command(["-V"])
        self.assertGreater(len(stdout), 0)
        self.assertEqual(stdout.split()[-1], kas.__version__)


class TestOutput(unittest.TestCase):
    """
    Tests that the output of the various tests is good.
    """

    def setUp(self):
        fd, self.temp_file = tempfile.mkstemp(prefix="htsget_cli_test_")
        os.close(fd)

    def tearDown(self):
        os.unlink(self.temp_file)

    def get_output(self, cmd):
        parser = cli.get_kastore_parser()
        args = parser.parse_args(cmd)
        return capture_output(args.runner, args)

    def get_example_data(self):
        data = {"A": np.arange(100), "B": np.zeros(10, dtype=int)}
        return data

    def test_list_empty(self):
        kas.dump({}, self.temp_file)
        stdout, stderr = self.get_output(["ls", self.temp_file])
        self.assertEqual(len(stderr), 0)
        self.assertEqual(len(stdout), 0)
        for opts in ["-l", "-lH"]:
            stdout, stderr = self.get_output(["ls", opts, self.temp_file])
            self.assertEqual(len(stderr), 0)
            self.assertEqual(len(stdout), 0)

    def test_list(self):
        data = self.get_example_data()
        kas.dump(data, self.temp_file)
        stdout, stderr = self.get_output(["ls", self.temp_file])
        self.assertEqual(len(stderr), 0)
        self.assertEqual(stdout.splitlines(), sorted(list(data.keys())))

    def test_list_long_format(self):
        data = self.get_example_data()
        kas.dump(data, self.temp_file)
        stdout, stderr = self.get_output(["ls", self.temp_file, "-l"])
        self.assertEqual(len(stderr), 0)
        lines = stdout.splitlines()
        self.assertEqual(len(lines), len(data))
        # TODO weak test, check the actual output.

    def test_list_long_format_human(self):
        data = self.get_example_data()
        kas.dump(data, self.temp_file)
        stdout, stderr = self.get_output(["ls", self.temp_file, "-l", "-H"])
        self.assertEqual(len(stderr), 0)
        lines = stdout.splitlines()
        self.assertEqual(len(lines), len(data))
        # TODO weak test, check the actual output.

    def test_dump(self):
        data = self.get_example_data()
        kas.dump(data, self.temp_file)
        for key in data.keys():
            stdout, stderr = self.get_output(["dump", self.temp_file, key])
            self.assertEqual(len(stderr), 0)
            self.assertEqual(stdout.splitlines(), list(map(str, data[key])))

    def test_error(self):
        # We really should be catching this exception and writing a message to
        # stderr. For now though, document the fact that we are raising the exception
        # up to the main.
        self.assertRaises(EOFError, self.get_output, ["ls", self.temp_file])

    def verify_logging(self, args, level):
        # We don't actually check the output here as we're mocking out the
        # call to logging config, but it's convenient to reuse the machinery
        # here in this class
        data = self.get_example_data()
        kas.dump(data, self.temp_file)
        log_format = "%(asctime)s %(message)s"
        with mock.patch("logging.basicConfig") as mocked_config:
            stdout, stderr = self.get_output(args + ["ls", self.temp_file])
            mocked_config.assert_called_once_with(level=level, format=log_format)
        return stderr

    def test_logging_verbosity_0(self):
        self.verify_logging([], logging.WARNING)

    def test_logging_verbosity_1(self):
        self.verify_logging(["-v"], logging.INFO)

    def test_logging_verbosity_2(self):
        self.verify_logging(["-vv"], logging.DEBUG)



import numpy
from setuptools import Extension
from setuptools import setup

_kastore_module = Extension(
    "_kastore",
    sources=["_kastoremodule.c", "lib/kastore.c"],
    extra_compile_args=["-std=c99"],
    include_dirs=["lib", numpy.get_include()],
)

setup(
    ext_modules=[_kastore_module],
)



import multiprocessing
import os
import pathlib
import platform
import queue
import shutil
import socket
import socketserver
import tempfile
import traceback
import unittest

import numpy as np

import kastore as kas

"""
Tests for load()ing and dump()ing with file-like objects.
"""


IS_WINDOWS = platform.system() == "Windows"


class DictVerifyMixin:
    def verify_dicts_equal(self, d1, d2):
        self.assertEqual(sorted(d1.keys()), sorted(d2.keys()))
        for key in d1.keys():
            np.testing.assert_equal(d1[key], d2[key])


# pathlib.Path objects should work  transparently, and these tests check it
# isn't broken by fileobj-enabling code.
class PathlibMixin(DictVerifyMixin):
    def setUp(self):
        fd, path = tempfile.mkstemp(prefix="kas_test_pathlib")
        os.close(fd)
        self.temp_file = pathlib.Path(path)

    def tearDown(self):
        self.temp_file.unlink()

    def test_dump_to_pathlib_Path(self):
        data = {"a": np.arange(10)}
        kas.dump(data, self.temp_file, engine=self.engine)

    def test_load_from_pathlib_Path(self):
        data = {"a": np.arange(10)}
        kas.dump(data, str(self.temp_file), engine=self.engine)
        file_size = self.temp_file.stat().st_size
        for read_all in [True, False]:
            data_out = kas.load(self.temp_file, read_all=read_all, engine=self.engine)
            data2 = dict(data_out.items())
            file_size2 = self.temp_file.stat().st_size
            self.verify_dicts_equal(data, data2)
            self.assertEqual(file_size, file_size2)


class FileobjMixin(DictVerifyMixin):
    def setUp(self):
        fd, path = tempfile.mkstemp(prefix="kas_test_fileobj")
        os.close(fd)
        self.temp_file = path

    def tearDown(self):
        os.unlink(self.temp_file)

    def test_dump_fileobj_single(self):
        data = {"a": np.arange(10)}
        with open(self.temp_file, "wb") as f:
            kas.dump(data, f, engine=self.engine)
        data_out = kas.load(self.temp_file, engine=self.engine)
        data2 = dict(data_out.items())
        self.verify_dicts_equal(data, data2)

    def test_dump_fileobj_multi(self):
        with open(self.temp_file, "wb") as f:
            for i in range(10):
                data = {
                    "i" + str(i): np.arange(i, dtype=int),
                    "f" + str(i): np.arange(i, dtype=float),
                }
                kas.dump(data, f, engine=self.engine)

    def test_load_fileobj_single(self):
        data = {"a": np.arange(10)}
        kas.dump(data, self.temp_file, engine=self.engine)
        file_size = os.stat(self.temp_file).st_size
        for read_all in [True, False]:
            with open(self.temp_file, "rb") as f:
                data_out = kas.load(f, read_all=read_all, engine=self.engine)
                data2 = dict(data_out.items())
                file_offset = f.tell()
            self.verify_dicts_equal(data, data2)
            self.assertEqual(file_offset, file_size)

    def test_load_and_dump_fileobj_single(self):
        data = {"a": np.arange(10)}
        with open(self.temp_file, "wb") as f:
            kas.dump(data, f, engine=self.engine)
        file_size = os.stat(self.temp_file).st_size
        for read_all in [True, False]:
            with open(self.temp_file, "rb") as f:
                data_out = kas.load(f, read_all=read_all, engine=self.engine)
                data2 = dict(data_out.items())
                file_offset = f.tell()
            self.verify_dicts_equal(data, data2)
            self.assertEqual(file_offset, file_size)

    def test_load_and_dump_fileobj_multi(self):
        datalist = [
            {
                "i" + str(i): i + np.arange(10**5, dtype=int),
                "f" + str(i): i + np.arange(10**5, dtype=float),
            }
            for i in range(10)
        ]
        file_offsets = []
        with open(self.temp_file, "wb") as f:
            for data in datalist:
                kas.dump(data, f, engine=self.engine)
                file_offsets.append(f.tell())
        for read_all in [True, False]:
            with open(self.temp_file, "rb") as f:
                for data, file_offset in zip(datalist, file_offsets):
                    data_out = kas.load(f, read_all=read_all, engine=self.engine)
                    data2 = dict(data_out.items())
                    file_offset2 = f.tell()
                    self.verify_dicts_equal(data, data2)
                    self.assertEqual(file_offset, file_offset2)

    def test_load_and_dump_file_single_rw(self):
        data = {"a": np.arange(10)}
        with open(self.temp_file, "r+b") as f:
            kas.dump(data, f, engine=self.engine)
            for read_all in [True, False]:
                f.seek(0)
                data_out = kas.load(f, read_all=read_all, engine=self.engine)
                data2 = dict(data_out.items())
                self.verify_dicts_equal(data, data2)

    def test_load_and_dump_file_multi_rw(self):
        datalist = [
            {
                "i" + str(i): i + np.arange(10**5, dtype=int),
                "f" + str(i): i + np.arange(10**5, dtype=float),
            }
            for i in range(10)
        ]
        with open(self.temp_file, "r+b") as f:
            for data in datalist:
                kas.dump(data, f, engine=self.engine)

            for read_all in [True, False]:
                f.seek(0)
                for data in datalist:
                    data_out = kas.load(f, read_all=read_all, engine=self.engine)
                    data2 = dict(data_out.items())
                    self.verify_dicts_equal(data, data2)

    def test_load_and_dump_fd_single_rw(self):
        data = {"a": np.arange(10)}
        with open(self.temp_file, "r+b") as f:
            fd = f.fileno()
            kas.dump(data, fd, engine=self.engine)
            for read_all in [True, False]:
                os.lseek(fd, 0, os.SEEK_SET)
                data_out = kas.load(fd, read_all=read_all, engine=self.engine)
                data2 = dict(data_out.items())
                self.verify_dicts_equal(data, data2)

    def test_load_and_dump_fd_multi_rw(self):
        datalist = [
            {
                "i" + str(i): i + np.arange(10**5, dtype=int),
                "f" + str(i): i + np.arange(10**5, dtype=float),
            }
            for i in range(20)
        ]
        with open(self.temp_file, "r+b") as f:
            fd = f.fileno()
            for data in datalist:
                kas.dump(data, fd, engine=self.engine)
            for read_all in [True, False]:
                os.lseek(fd, 0, os.SEEK_SET)
                for data in datalist:
                    data_out = kas.load(fd, read_all=read_all, engine=self.engine)
                    data2 = dict(data_out.items())
                    self.verify_dicts_equal(data, data2)


def dump_to_stream(q_err, q_in, file_out, engine):
    """
    Get data dicts from `q_in` and kas.dump() them to `file_out`.
    Uncaught exceptions are placed onto the `q_err` queue.
    """
    try:
        with open(file_out, "wb") as f:
            while True:
                data = q_in.get()
                if data is None:
                    break
                kas.dump(data, f, engine=engine)
    except Exception as exc:
        tb = traceback.format_exc()
        q_err.put((exc, tb))


def load_from_stream(q_err, q_out, file_in, engine, read_all):
    """
    kas.load() stores from `file_in` and put their data onto `q_out`.
    Uncaught exceptions are placed onto the `q_err` queue.
    """
    try:
        with open(file_in, "rb") as f:
            while True:
                try:
                    data = kas.load(f, read_all=read_all, engine=engine)
                except EOFError:
                    break
                q_out.put(dict(data.items()))
    except Exception as exc:
        tb = traceback.format_exc()
        q_err.put((exc, tb))


class StreamingMixin(DictVerifyMixin):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp(prefix="kas_test_streaming")
        self.temp_fifo = os.path.join(self.temp_dir, "fifo")
        os.mkfifo(self.temp_fifo)

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def stream(self, datalist, read_all=True):
        """
        data -> q_in -> kas.dump(..., fifo) -> kas.load(fifo) -> q_out -> data_out
        """
        q_err = multiprocessing.Queue()
        q_in = multiprocessing.Queue()
        q_out = multiprocessing.Queue()
        proc1 = multiprocessing.Process(
            target=dump_to_stream, args=(q_err, q_in, self.temp_fifo, self.engine)
        )
        proc2 = multiprocessing.Process(
            target=load_from_stream,
            args=(q_err, q_out, self.temp_fifo, self.engine, read_all),
        )
        proc1.start()
        proc2.start()
        for data in datalist:
            q_in.put(data)
        q_in.put(None)  # signal the process that we're done
        proc1.join(timeout=3)
        if not q_err.empty():
            # re-raise the first child exception
            exc, tb = q_err.get()
            print(tb)
            raise exc
        if proc1.is_alive():
            # prevent hang if proc1 failed to join
            proc1.terminate()
            proc2.terminate()
            self.assertTrue(False, msg="proc1 (kas.dump) failed to join")
        datalist_out = []
        for _ in datalist:
            try:
                data_out = q_out.get(timeout=3)
            except queue.Empty:
                # terminate proc2 so we don't hang
                proc2.terminate()
                raise
            datalist_out.append(data_out)
        proc2.join(timeout=3)
        if proc2.is_alive():
            # prevent hang if proc2 failed to join
            proc2.terminate()
            self.assertTrue(False, msg="proc2 (kas.load) failed to join")

        self.assertEqual(len(datalist), len(datalist_out))
        for data, data_out in zip(datalist, datalist_out):
            self.verify_dicts_equal(data, data_out)

    def test_stream_single(self):
        datalist = [{"a": np.array([0])}]
        self.stream(datalist)

    def test_stream_multi(self):
        datalist = [
            {
                "i" + str(i): i + np.arange(10**5, dtype=int),
                "f" + str(i): i + np.arange(10**5, dtype=float),
            }
            for i in range(100)
        ]
        self.stream(datalist)


ADDRESS = ("localhost", 10009)


class TestServer(socketserver.ThreadingTCPServer):
    allow_reuse_address = True


class StoreEchoHandler(socketserver.BaseRequestHandler):
    def handle(self):
        while True:
            try:
                data = kas.load(
                    self.request.fileno(), engine=self.engine, read_all=True
                )
            except EOFError:
                break
            kas.dump(dict(data), self.request.fileno(), engine=self.engine)
        # We only read one list, so shutdown the server straight away
        self.server.shutdown()


class StoreEchoHandlerCEngine(StoreEchoHandler):
    engine = kas.C_ENGINE


class StoreEchoHandlerPyEngine(StoreEchoHandler):
    engine = kas.PY_ENGINE


def server_process(engine, q):
    handlers = {
        kas.C_ENGINE: StoreEchoHandlerCEngine,
        kas.PY_ENGINE: StoreEchoHandlerPyEngine,
    }
    server = TestServer(ADDRESS, handlers[engine])
    # Tell the client (on the other end of the queue) that it's OK to open
    # a connection
    q.put(None)
    server.serve_forever()


class SocketMixin(DictVerifyMixin):
    def setUp(self):
        # Use a queue to synchronise the startup of the server and the client.
        q = multiprocessing.Queue()
        self.server_process = multiprocessing.Process(
            target=server_process, args=(self.engine, q)
        )
        self.server_process.start()
        q.get()
        self.client = socket.create_connection(ADDRESS)

    def tearDown(self):
        self.client.close()
        self.server_process.join()

    def verify_stream(self, data_list):
        fd = self.client.fileno()
        for data in data_list:
            kas.dump(data, fd, engine=self.engine)
            echo_data = kas.load(fd, read_all=True, engine=self.engine)
            self.verify_dicts_equal(data, echo_data)

    def test_single(self):
        self.verify_stream([{"a": np.arange(10)}])

    def test_two(self):
        self.verify_stream([{"a": np.zeros(10)}, {"b": np.zeros(100)}])

    def test_multi(self):
        datalist = [
            {
                "i" + str(i): i + np.arange(10**5, dtype=int),
                "f" + str(i): i + np.arange(10**5, dtype=float),
            }
            for i in range(10)
        ]
        self.verify_stream(datalist)


class TestPathlibCEngine(PathlibMixin, unittest.TestCase):
    engine = kas.C_ENGINE


class TestPathlibPyEngine(PathlibMixin, unittest.TestCase):
    engine = kas.PY_ENGINE


class TestFileobjCEngine(FileobjMixin, unittest.TestCase):
    engine = kas.C_ENGINE


class TestFileobjPyEngine(FileobjMixin, unittest.TestCase):
    engine = kas.PY_ENGINE


@unittest.skipIf(IS_WINDOWS, "FIFOs don't exist on Windows")
class TestStreamingCEngine(StreamingMixin, unittest.TestCase):
    engine = kas.C_ENGINE


@unittest.skipIf(IS_WINDOWS, "FIFOs don't exist on Windows")
class TestStreamingPyEngine(StreamingMixin, unittest.TestCase):
    engine = kas.PY_ENGINE


@unittest.skipIf(IS_WINDOWS, "Deadlocking on Windows")
class TestSocketCEngine(SocketMixin, unittest.TestCase):
    engine = kas.C_ENGINE


@unittest.skipIf(IS_WINDOWS, "Deadlocking on Windows")
class TestSocketPyEngine(SocketMixin, unittest.TestCase):
    engine = kas.PY_ENGINE
